/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

/* Cherokee
 *
 * Authors:
 *      Alvaro Lopez Ortega <alvaro@alobbs.com>
 *
 * Copyright (C) 2001-2008 Alvaro Lopez Ortega
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include "common-internal.h"
#include <signal.h>
#include "init.h"
#include "server.h"

#ifdef HAVE_GETOPT_H
# include <getopt.h>
#endif

/* Notices 
 */
#define APP_NAME        \
	"Cherokee Web Server"

#define APP_COPY_NOTICE \
	"Written by Alvaro Lopez Ortega <alvaro@gnu.org>\n\n"                          \
	"Copyright (C) 2001-2008 Alvaro Lopez Ortega.\n"                               \
	"This is free software; see the source for copying conditions.  There is NO\n" \
	"warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"


/* Default configuration
 */
#define DEFAULT_CONFIG_FILE CHEROKEE_CONFDIR "/cherokee.conf"

#define BASIC_CONFIG                                                      \
	"vserver!default!rule!1!match = default\n"                        \
	"vserver!default!rule!1!handler = common\n"                       \
	"vserver!default!rule!1!handler!iocache = 0\n"                    \
	"vserver!default!rule!2!match = directory\n"                      \
	"vserver!default!rule!2!match!directory = /icons\n"               \
	"vserver!default!rule!2!handler = file\n"                         \
	"vserver!default!rule!2!document_root = " CHEROKEE_ICONSDIR "\n"  \
	"vserver!default!rule!3!match = directory\n"                      \
	"vserver!default!rule!3!match!directory = /cherokee_themes\n"     \
	"vserver!default!rule!3!handler = file\n"                         \
	"vserver!default!rule!3!document_root = " CHEROKEE_THEMEDIR "\n"  \
	"icons!default = page_white.png\n"                                \
	"icons!directory = folder.png\n"      	                          \
	"icons!parent_directory = arrow_turn_left.png\n"                  \
	"try_include = " CHEROKEE_CONFDIR "/mods-enabled\n"

static cherokee_server_t  *srv           = NULL;
static char               *config_file   = NULL;
static char               *document_root = NULL;
static cherokee_boolean_t  daemon_mode   = false;
static cherokee_boolean_t  just_test     = false;
static cuint_t             port          = 80;

static ret_t common_server_initialization (cherokee_server_t *srv);


static void
panic_handler (int code)
{
	UNUSED(code);
	cherokee_server_handle_panic (srv);
}

static void
prepare_to_die (int code)
{
	UNUSED(code);
	cherokee_server_handle_TERM (srv);
}

static void
restart_server_cb (cherokee_server_t *new_srv)
{
	ret_t ret;

	srv = new_srv;
	ret = common_server_initialization (srv);
	if (ret != ret_ok) exit(3);
}

static void
restart_server (int code)
{	
	UNUSED(code);
	printf ("Handling HUP signal..\n");
	cherokee_server_handle_HUP (srv, restart_server_cb);
}

static void
reopen_log_files (int code)
{	
	UNUSED(code);
	printf ("Reopeing log files..\n");
	cherokee_server_log_reopen (srv);
}

static ret_t
test_configuration_file (void)
{
	ret_t   ret;
	char   *config;

	config = (config_file) ? config_file : DEFAULT_CONFIG_FILE;
	ret = cherokee_server_read_config_file (srv, config);

	PRINT_MSG ("Test on %s: %s\n", config, (ret == ret_ok)? "OK": "Failed"); 
	return ret;
}

static ret_t
common_server_initialization (cherokee_server_t *srv)
{
	ret_t ret;

#ifdef SIGPIPE
        signal (SIGPIPE, SIG_IGN);
#endif
#ifdef SIGHUP
        signal (SIGHUP, restart_server);
#endif
#ifdef SIGUSR1
        signal (SIGUSR1, reopen_log_files);
#endif
#ifdef SIGSEGV
        signal (SIGSEGV, panic_handler);
#endif
#ifdef SIGTERM
        signal (SIGTERM, prepare_to_die);
#endif

	if (document_root != NULL) {
		cherokee_buffer_t tmp = CHEROKEE_BUF_INIT;

		cherokee_buffer_add_va (&tmp, 
					"server!port = %d\n"
					"vserver!default!document_root = %s\n"
					BASIC_CONFIG, port, document_root);

		ret = cherokee_server_read_config_string (srv, &tmp);
		cherokee_buffer_mrproper (&tmp);

		if (ret != ret_ok) {
			PRINT_MSG ("Couldn't start serving directory %s\n", document_root);
			return ret_error;
		}

	} else {
		char *config;

		config = (config_file) ? config_file : DEFAULT_CONFIG_FILE;
		ret = cherokee_server_read_config_file (srv, config);

		if (ret != ret_ok) {
			PRINT_MSG ("Couldn't read the config file: %s\n", config);
			return ret_error;
		}
	}

	if (daemon_mode)
		cherokee_server_daemonize (srv);

	cherokee_server_write_pidfile (srv);

	ret = cherokee_server_initialize (srv);
	if (ret != ret_ok) return ret_error;

	cherokee_server_unlock_threads (srv);
	return ret_ok;
}

static void
print_help (void)
{
	printf (APP_NAME "\n"
		"Usage: cherokee [options]\n\n"
		"  -h,  --help                   Print this help\n"
		"  -V,  --version                Print version and exit\n"
		"  -t,  --test                   Just test configuration file\n"
		"  -d,  --detach                 Detach from the console\n"
		"  -C,  --config=PATH            Configuration file\n"
		"  -p,  --port=NUM               TCP port number\n"
		"  -r,  --documentroot=PATH      Server directory content\n\n"
		"Report bugs to " PACKAGE_BUGREPORT "\n");
}

static void
process_parameters (int argc, char **argv)
{
	int c;

	struct option long_options[] = {
		{"help",         no_argument,       NULL, 'h'},
		{"version",      no_argument,       NULL, 'V'},
		{"detach",       no_argument,       NULL, 'd'},
		{"test",         no_argument,       NULL, 't'},
		{"port",         required_argument, NULL, 'p'},
		{"documentroot", required_argument, NULL, 'r'},
		{"config",       required_argument, NULL, 'C'},
		{NULL, 0, NULL, 0}
	};

	while ((c = getopt_long(argc, argv, "hVdtp:r:C:", long_options, NULL)) != -1) {
		switch(c) {
		case 'C':
			config_file = strdup(optarg);
			break;
		case 'd':
			daemon_mode = true;
			break;
		case 'r':
			document_root = strdup(optarg);
			break;
		case 'p':
			port = atoi(optarg);
			break;
		case 't':
			just_test = true;
			break;
		case 'V':
			printf (APP_NAME " " PACKAGE_VERSION "\n" APP_COPY_NOTICE);
			exit(0);
		case 'h':
		case '?':
		default:
			print_help();
			exit(0);
		}
	}
}


int
main (int argc, char **argv)
{
	ret_t ret;

	cherokee_init();

	ret = cherokee_server_new (&srv);
	if (ret < ret_ok) return 1;
	
	process_parameters (argc, argv);
	
	if (just_test) {
		ret = test_configuration_file();
		exit (ret);
	}

	ret = common_server_initialization (srv);
	if (ret < ret_ok) return 2;

	do {
		ret = cherokee_server_step (srv);
	} while (ret == ret_eagain);
	
	cherokee_server_stop (srv);
	cherokee_server_free (srv);

	if (config_file)
		free (config_file);

	return 0;
}
